/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2023 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"

#include "globals.hh"
#include "utils.hh"
#include "support_utils.hh"
#include "Bits.hh"
#include "Con.hh"
#include "Gen.hh"
#include "Index_Partition.hh"
#include "Integer.hh"
#include "Integer.hh"
#include "Linear_Expr.hh"
#include "Poly.hh"
#include "Poly_templ.hh"
#include "Poly_min.hh"
#include "Poly_Rel.hh"
#include "Sat.hh"
#include "Scalar_Prod.hh"

#include <algorithm>
#include <cassert>
#include <utility>
#include <vector>

namespace pplite {
namespace detail {

using Split_Result = std::pair<Conv_Iter_Result, Conv_Iter_Result>;

inline bool
has_no_eq_point(const Ranges& sk_ranges,
                const Gens& sk, const NS_Rows& ns_rows) {
  Index_Set sk_eq = sk_ranges_to_eq_index_set(sk_ranges);
  for (const auto& g : sk_eq)
    if (sk[g].is_point())
      return false;
  return !any_of(ns_rows, [&sk_eq](const Index_Set& ns) {
                            return subset_eq(ns, sk_eq);
                          });
}

inline bool
has_no_eq_point(const Ranges& sk_ranges, const Gens& sk) {
  NS_Rows dummy;
  return has_no_eq_point(sk_ranges, sk, dummy);
}

/*******************************************************/
// Process skeleton

// Combines the skeleton generator in SK+ with those in SK-.
// Creates new elements in sk_strict and, if nnc_split, also in sk_nonstrict.
// Updates only the `sat_strict': its copy is postponed.
void
split_combine_sk_rows(Ranges& sk_ranges, Gens& sk_strict, Gens& sk_nonstrict,
                      Integers& sp, Sat& sat_strict,
                      const bool nnc_split,
                      const dim_type min_sat, const dim_type max_sat) {
  const dim_type old_num_sk_rows = sk_strict.size();

  static PPLITE_TLS Dims sk_pos, sk_neg;
  static PPLITE_TLS Dims sat_ones;
  sk_pos.clear();
  sk_neg.clear();
  sat_ones.clear();
  for (auto i : index_range(sk_ranges)) {
    if (sk_ranges[i] == Range::EQ) {
      sat_ones.push_back(0);
      continue;
    }
    sat_ones.push_back(sat_strict[i].count_ones());
    if (sk_ranges[i] == Range::POS)
      sk_pos.push_back(i);
    else
      sk_neg.push_back(i);
  }

  for (auto pos : sk_pos) {
    for (auto neg : sk_neg) {
      // Exploit saturation information to perform adjacency tests.
      // Note: delay creation of `new_row' and `new_satrow'.
      const dim_type new_satrow_ones
        = sat_strict[pos].count_ones_in_union(sat_strict[neg]);
      if (quick_non_adj_test(min_sat, max_sat, new_satrow_ones))
        continue;
      Bits new_satrow = Bits::get_union(sat_strict[pos], sat_strict[neg]);
      if (not quick_adj_test(pos, neg, new_satrow_ones, sat_ones) &&
          not combinatorial_adj_test(0, old_num_sk_rows,
                                     pos, neg, new_satrow, sat_strict))
        continue;

      // Here `pos' and `neg' are adjacent: update `sat_strict'.
      sat_strict.add_row(std::move(new_satrow));

      // Add the new row to `sk_strict'
      auto new_row_strict = sk_strict[neg];
      // Combination can be done once.
      combine_row(new_row_strict, sk_strict[pos], sp[neg], sp[pos]);
      new_row_strict.set_type(combine_type(sk_strict[pos].type(),
                                           sk_strict[neg].type(),
                                           nnc_split));
      if (nnc_split) {
        // Copy the new element
        auto new_row_nonstrict = new_row_strict;
        // Set the right type, according to the side.
        new_row_nonstrict.set_type(combine_type(sk_strict[neg].type(),
                                                sk_strict[pos].type(),
                                                false));
        sk_nonstrict.push_back(std::move(new_row_nonstrict));
      }
      sk_strict.push_back(std::move(new_row_strict));

      assert(num_rows(sk_strict) == sat_strict.num_rows());
    } // end loop on sk_neg
  } // end loop on sk_pos

  sp.resize(sk_strict.size());
  sk_ranges.resize(sk_strict.size(), Range::EQ);
}


/**
 * Computes the split sk systems, copying the original one.
 * NB: sk_[non]strict maskes sense iff [non]strict result is not EMPTY.
 * If a component is EMPTY, it will be the caller to set the empty DD.
 */
Split_Result
Q_split_skel(dim_type space_dim, dim_type num_lines, dim_type num_eq,
             Gens& sk_strict, Gens& sk_nonstrict,
             Ranges& sk_ranges, const Con& beta_strict,
             const NS_Rows& ns_strict, Sat& sat_strict) {
  assert(beta_strict.is_inequality());
  const bool nnc_split = beta_strict.is_strict_inequality();
  bool empty_strict = false;
  bool empty_nonstrict = false;

  Integers scal_prods;
  sp::compute_prods(beta_strict, sk_strict, scal_prods);

  // Create partition for skeleton rows.
  // The partition is relative to the strict side.
  sk_partition(scal_prods, sk_ranges);
  Index_Partition<Range::NEG> sk_neg(sk_ranges);
  Index_Partition<Range::EQ> sk_eq(sk_ranges);
  Index_Partition<Range::POS> sk_pos(sk_ranges);

  if (!sk_neg.empty() && !sk_pos.empty()) {
    // Here SK+ and SK- are non-empty.
    // Combine SK+ and SK- *adjacent* generators.
    const dim_type min_sat = space_dim - num_lines - 1;
    const dim_type max_sat = sat_strict.num_cols() + num_eq;
    if (nnc_split) {
      // Copy sk_strict into sk_nonstrict
      sk_nonstrict = sk_strict;
    }
    split_combine_sk_rows(sk_ranges, sk_strict, sk_nonstrict,
                          scal_prods, sat_strict,
                          nnc_split, min_sat, max_sat);
  } else {
    // Here SK+ or SK- is empty.
    const bool no_eq_point = has_no_eq_point(sk_ranges, sk_strict, ns_strict);
    if (sk_neg.empty()) {
      if (sk_eq.empty() || no_eq_point)
        empty_nonstrict = true;
      if (nnc_split && sk_pos.empty()) {
        // strict side is only in eq, but it is strict.
        empty_strict = true;
      }
      // The non-skel handling is postponed.
    }
    else {
      assert(sk_pos.empty());
      if (nnc_split || sk_eq.empty() || no_eq_point)
        empty_strict = true;
    }

    // Copy original gs system only if both sides are non-empty.
    // If !nnc_split the copy is postponed, since there could be
    // shared promotions of the same ns elements.
    if (!empty_strict && !empty_nonstrict && nnc_split)
      sk_nonstrict = sk_strict;
  }

  assert(not (empty_strict && empty_nonstrict));
  if (empty_strict)
    return { Conv_Iter_Result::EMPTY, Conv_Iter_Result::REDUNDANT };
  if (empty_nonstrict)
    return { Conv_Iter_Result::REDUNDANT, Conv_Iter_Result::EMPTY };

  // Update sat_strict cols.
  sat_strict.add_cols(1);
  update_last_sat_column(sat_strict, sk_pos);
  update_last_sat_column(sat_strict, sk_neg);
  return { Conv_Iter_Result::OK, Conv_Iter_Result::OK };
}

/*******************************************************/
// Process non-skeleton

/* Move-ns method for NNC split */
// Computes the closures once and differentiates only the projection phase.
// Checks the supports to be promoted only in the nonstrict side.
void
split_move_ns(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
              Ranges& ns_ranges_strict, Ranges& ns_ranges_nonstrict,
              Gens& sk_nonstrict,
              const Index_Set& iset_neg,
              const Index_Set& iset_eq,
              const Index_Set& iset_pos,
              NS_Rows& nsstar_strict, NS_Rows& nsstar_nonstrict,
              const Sat& sat_c, Sat& sat_g) {
  // Originally ns_ranges are the same.
  Dims to_be_del;
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  for (auto i : ns_pos_neg) {
    assert(ns_ranges_strict[i] == Range::POS_NEG);
    Index_Set& supp_strict = ns_strict[i];
    // Checkme: closures can be done once, since sat info are shared.
    support_closure(supp_strict, sat_c, sat_g);

    Index_Set supp_nonstrict = supp_strict;
    // Differentiate only the projection phase.
    supp_projection(supp_strict, true, iset_neg, iset_eq);
    supp_projection(supp_nonstrict, false, iset_pos, iset_eq);
    // No need to check if it has to be promoted
    assert(supp_strict.size() > 1);
    add_to_minimal_set_of_ns(supp_strict, nsstar_strict);

    // Check if it can be moved into SK.
    if (supp_nonstrict.size() == 1) {
      // Change its only supporter's type into point.
      const auto supporter_index = *(supp_nonstrict.begin());
      promote_singleton(sk_nonstrict[supporter_index]);
      to_be_del.push_back(i);
      // Consider next ns in pos_neg
      continue;
    }
    assert(supp_nonstrict.size() > 1);
    add_to_minimal_set_of_ns(std::move(supp_nonstrict), nsstar_nonstrict);
  }
  // Remove ns that has been moved into SK.
  erase_using_sorted_indices(ns_nonstrict, to_be_del);
  erase_using_sorted_indices(ns_ranges_nonstrict, to_be_del);
}

/** Create_ns method for NNC split. **/
// Add to NS*strict:
// - ns* from NEG projected in POS,
// - ns* from EQ projected in POS.
// Add to NS*nonstrict:
// - ns* from NEG projected in EQ,
// - ns* from POS projected in EQ.
void
split_create_ns(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
                Ranges& ns_ranges_strict, Ranges& ns_ranges_nonstrict,
                Gens& sk_strict, Gens& sk_nonstrict, const Ranges& sk_ranges,
                const Index_Set& iset_pos,
                const Index_Set& iset_eq,
                const Index_Set& iset_neg,
                NS_Rows& nsstar_strict, NS_Rows& nsstar_nonstrict,
                const Sat& sat_c, const Sat& sat_g,
                dim_type min_sat, dim_type max_sat,
                bool should_create_nonstrict) {
  // enumerate_faces from NEG can be done once:
  // Store the computed set, reproject it in EQ if needed.
  NS_Rows set_from_neg;
  create_ns_from_Q_with_face_enum<Range::NEG>(Range::POS,
                                              sk_strict, ns_strict,
                                              sk_ranges, ns_ranges_strict,
                                              set_from_neg,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              true);
  // Here `set_from_neg' stores the created ns projected in POS.
  for (auto& ns_star : set_from_neg) {
    // Add it to nsstar_strict (it is copied).
    add_to_minimal_set_of_ns(ns_star, nsstar_strict);
    if (should_create_nonstrict) {
      // Reproject it in EQ
      supp_projection(ns_star, false, iset_neg, iset_eq);
      // Add it to nsstar_nonstrict too.
      add_to_minimal_set_of_ns(std::move(ns_star), nsstar_nonstrict);
    }
  }

  // Add to nsstar_strict new ns from EQ.
  create_ns_from_eq_points(sk_strict, ns_strict, sk_ranges, ns_ranges_strict,
                           nsstar_strict, iset_pos, iset_eq, iset_neg,
                           sat_c, sat_g, min_sat, max_sat);

  if (should_create_nonstrict) {
    // Add to nsstar_nonstrict new ns from POS.
    create_ns_from_Q_with_face_enum<Range::POS>(Range::NEG,
                                                sk_nonstrict, ns_nonstrict,
                                                sk_ranges, ns_ranges_nonstrict,
                                                nsstar_nonstrict,
                                                iset_pos, iset_eq, iset_neg,
                                                sat_c, sat_g,
                                                false);
  }
}

void
split_non_skel(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
               Gens& sk_strict, Gens& sk_nonstrict,
               const Ranges& sk_ranges,
               const Sat& sat_c, Sat& sat_g,
               dim_type space_dim, dim_type num_lines, dim_type num_eq) {
  NS_Rows nsstar_strict, nsstar_nonstrict;
  Ranges ns_ranges_strict, ns_ranges_nonstrict;
  // Create partition for non-skeleton rows.
  Index_Set iset_neg, iset_eq, iset_pos;
  sk_ranges_to_index_sets(sk_ranges, iset_neg, iset_eq, iset_pos);
  ns_partition(ns_strict, iset_neg, iset_pos, ns_ranges_strict);

  // Copy partition, since they will follow different ns elements.
  // NB: the real nonstrict partition should be its opposite,
  // it is not complemented for efficiency.
  ns_nonstrict = ns_strict;
  ns_ranges_nonstrict = ns_ranges_strict;
  nsstar_strict.clear();
  nsstar_nonstrict.clear();

  bool sat_is_up_to_date = false;
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  if (!ns_pos_neg.empty()) {
    sat_g = sat_c.transpose();
    sat_is_up_to_date = true;
    split_move_ns(ns_strict, ns_nonstrict,
                  ns_ranges_strict, ns_ranges_nonstrict,
                  sk_nonstrict,
                  iset_neg, iset_eq, iset_pos,
                  nsstar_strict, nsstar_nonstrict,
                  sat_c, sat_g);
  }

  // Check if we need to call `create_ns' for the non-strict side too.
  const bool should_create_nonstrict =
    can_have_nonskel(iset_eq, sk_nonstrict);
  if (!sat_is_up_to_date) {
    sat_g = sat_c.transpose();
    sat_is_up_to_date = true;
  }
  const dim_type min_sat = space_dim - num_eq - 1;
  const dim_type max_sat = sat_c.num_cols() + num_lines;
  split_create_ns(ns_strict, ns_nonstrict,
                  ns_ranges_strict, ns_ranges_nonstrict,
                  sk_strict, sk_nonstrict, sk_ranges,
                  iset_pos, iset_eq, iset_neg,
                  nsstar_strict, nsstar_nonstrict,
                  sat_c, sat_g, min_sat, max_sat,
                  should_create_nonstrict);
  // Add new ns.
  if (sat_is_up_to_date) {
    new_ns_from_set(sk_strict, ns_strict, nsstar_strict,
                    sk_ranges, ns_ranges_strict, true);
    new_ns_from_set(sk_nonstrict, ns_nonstrict, nsstar_nonstrict,
                    sk_ranges, ns_ranges_nonstrict, false);
  }

  // Final update
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::POS_NEG, Range::NEG);
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::EQ, Range::NEG);
  std::replace(ns_ranges_nonstrict.begin(), ns_ranges_nonstrict.end(),
               Range::POS_NEG, Range::POS);
  Index_Partition<Range::NEG> ns_neg(ns_ranges_strict);
  Index_Partition<Range::POS> ns_pos(ns_ranges_nonstrict);
  erase_using_sorted_indices(ns_strict, ns_neg);
  erase_using_sorted_indices(ns_nonstrict, ns_pos);
}

/*******************************************************/
// Non-skel handling in CLOSED split.

/* Overloaded move-ns method for CLOSED split */
// Compute the closures and project in EQ.
// Check the supports to be promoted.
void
split_move_ns(NS_Rows& ns_rows, Ranges& ns_ranges, Gens& sk_rows,
              const Index_Set& iset_neg,
              const Index_Set& iset_eq,
              NS_Rows& nsstar,
              const Sat& sat_c, Sat& sat_g) {
  // Originally ns_ranges are the same.
  Dims to_be_del;
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges);
  for (auto i : ns_pos_neg) {
    assert(ns_ranges[i] == Range::POS_NEG);
    Index_Set& supp = ns_rows[i];
    support_closure(supp, sat_c, sat_g);
    supp_projection(supp, false, iset_neg, iset_eq);

    // Check if it can be moved into SK.
    if (supp.size() == 1) {
      // Change its only supporter's type into point.
      const auto supporter_index = *(supp.begin());
      promote_singleton(sk_rows[supporter_index]);
      to_be_del.push_back(i);
      // Consider next ns in pos_neg
      continue;
    }
    add_to_minimal_set_of_ns(supp, nsstar);
  }
  // Remove ns that has been moved into SK.
  erase_using_sorted_indices(ns_rows, to_be_del);
  erase_using_sorted_indices(ns_ranges, to_be_del);
}

/** Overloaded create_ns method for CLOSED split. **/
// Add to NS*:
// - ns* from NEG projected in EQ,
// - ns* from POS projected in EQ.
void
split_create_ns(NS_Rows& ns_rows, Ranges& ns_ranges,
                Gens& sk_rows, const Ranges& sk_ranges,
                const Index_Set& iset_pos,
                const Index_Set& iset_eq,
                const Index_Set& iset_neg,
                NS_Rows& nsstar,
                const Sat& sat_c, const Sat& sat_g) {
  create_ns_from_Q_with_face_enum<Range::NEG>(Range::POS,
                                              sk_rows, ns_rows,
                                              sk_ranges, ns_ranges,
                                              nsstar,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              false);
  create_ns_from_Q_with_face_enum<Range::POS>(Range::NEG,
                                              sk_rows, ns_rows,
                                              sk_ranges, ns_ranges,
                                              nsstar,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              false);
}

/* Overloaded version for CLOSED split */
// Computes only one NS* and copies the whole NS sys only in the end.
void
split_non_skel(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
               Gens& sk_rows, const Ranges& sk_ranges,
               const Sat& sat_c, Sat& sat_g) {
  NS_Rows nsstar_strict;
  Ranges ns_ranges_strict;
  // Create partition for non-skeleton rows.
  Index_Set iset_neg, iset_eq, iset_pos;
  sk_ranges_to_index_sets(sk_ranges, iset_neg, iset_eq, iset_pos);
  ns_partition(ns_strict, iset_neg, iset_pos, ns_ranges_strict);
  nsstar_strict.clear();

  bool sat_is_up_to_date = false;
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  if (!ns_pos_neg.empty()) {
    sat_g = sat_c.transpose();
    sat_is_up_to_date = true;
    split_move_ns(ns_strict, ns_ranges_strict, sk_rows,
                  iset_neg, iset_eq,
                  nsstar_strict, sat_c, sat_g);
  }

  // Check if we need to call `create_ns'.
  const bool should_create_nonstrict =
    can_have_nonskel(iset_eq, sk_rows);
  if (should_create_nonstrict) {
    if (!sat_is_up_to_date) {
      sat_g = sat_c.transpose();
      sat_is_up_to_date = true;
    }
    split_create_ns(ns_strict, ns_ranges_strict,
                    sk_rows, sk_ranges,
                    iset_pos, iset_eq, iset_neg,
                    nsstar_strict,
                    sat_c, sat_g);
  }
  // Add new ns.
  if (sat_is_up_to_date) {
    new_ns_from_set(sk_rows, ns_strict, nsstar_strict,
                    sk_ranges, ns_ranges_strict, true);
  }

  // Copy NS systems.
  Ranges ns_ranges_nonstrict = ns_ranges_strict;
  ns_nonstrict = ns_strict;

  // Final update
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::POS_NEG, Range::NEG);
  std::replace(ns_ranges_nonstrict.begin(), ns_ranges_nonstrict.end(),
               Range::POS_NEG, Range::POS);
  Index_Partition<Range::NEG> ns_neg(ns_ranges_strict);
  Index_Partition<Range::POS> ns_pos(ns_ranges_nonstrict);
  erase_using_sorted_indices(ns_strict, ns_neg);
  erase_using_sorted_indices(ns_nonstrict, ns_pos);
}

/*******************************************************/
// Process violating line

void
split_violating_singular(dim_type ex,
                         Gen_Sys& sys_strict,
                         Gen_Sys& sys_nonstrict,
                         Sat& sat_c_strict, Sat& sat_c_nonstrict,
                         const Con& beta_strict) {
  auto& sg_strict = sys_strict.sing_rows;
  auto& sk_strict = sys_strict.sk_rows;
  auto& ns_strict = sys_strict.ns_rows;
  auto& sk_nonstrict = sys_nonstrict.sk_rows;

  const bool nnc_split = beta_strict.is_strict_inequality();

  // Map the ex_strict in a ray
  auto& ex_strict = sg_strict[ex];
  Integer sp_ex;
  sp::assign(sp_ex, beta_strict, ex_strict);
  make_non_singular(ex_strict, sp_ex);
  // Map all other geometric rows into SK=: can be done once.
  combine_with_ex_sing(ex, sp_ex, beta_strict, sg_strict, sk_strict);
  // Move the ex singular to sk
  const auto new_ex = num_rows(sk_strict);
  sk_strict.push_back(std::move(ex_strict));
  sg_strict.erase(sg_strict.begin() + ex);
  // Copy sys: change ray orientation for the non-strict side.
  sys_nonstrict = sys_strict;
  auto& ex_nonstrict = sk_nonstrict[num_rows(sk_nonstrict) - 1];
  assert(ex_nonstrict == sk_strict[new_ex]);
  neg_assign(ex_nonstrict.linear_expr());

  // Update sat_c : can be done once.
  assert(beta_strict.is_inequality());
  sat_c_strict.add_cols(1);
  Bits sat_ex;
  sat_ex.set(sat_c_strict.num_cols() - 1);
  sat_c_strict.add_row(std::move(sat_ex));
  // Copy sat: sat info are shared.
  sat_c_nonstrict = sat_c_strict;

  // Here sk_nonstrict and ns_nonstrict are done.

  if (nnc_split) {
    // Handle ns_strict and sk_strict eq points.
    Ranges sk_ranges, ns_ranges;
    sk_ranges.assign(sk_strict.size(), Range::EQ);
    sk_ranges[new_ex] = Range::POS;
    ns_ranges.assign(ns_strict.size(), Range::EQ);
    create_ns_with_ex_sing(sk_strict, ns_strict,
                           sk_ranges, ns_ranges, new_ex);
  }
}

/********************************************************/

/**
 * Split with RATIONAL VARIABLES.
 * This is the usual conversion algorithm, only
 * updating gen_sys_then and gen_sys_else differently.
 */
Split_Result
Q_split_gen_sys(dim_type space_dim, dim_type num_eq,
                Gen_Sys& gen_sys_strict, Gen_Sys& gen_sys_nonstrict,
                Sat& sat_c_strict, Sat& sat_c_nonstrict,
                Sat& sat_g_strict,
                const Con& beta_strict, bool nnc_poly) {
  // If `nnc_split' then the "strict" side is actually strict.
  const bool nnc_split = beta_strict.is_strict_inequality();
  assert(!nnc_split || nnc_poly);

  Gens& sg_strict = gen_sys_strict.sing_rows;
  Gens& sk_strict = gen_sys_strict.sk_rows;
  Gens& sk_nonstrict = gen_sys_nonstrict.sk_rows;
  NS_Rows& ns_strict = gen_sys_strict.ns_rows;
  NS_Rows& ns_nonstrict = gen_sys_nonstrict.ns_rows;

  // Check if all singular rows saturate strict_beta.
  const dim_type ex = check_sing_rows(beta_strict, sg_strict);
  if (ex < num_rows(sg_strict)) {
    split_violating_singular(ex, gen_sys_strict,
                             gen_sys_nonstrict,
                             sat_c_strict, sat_c_nonstrict,
                             beta_strict);
    return { Conv_Iter_Result::OK, Conv_Iter_Result::OK };
  }

  // Here all lines saturate the constraint.
  // Copy sing sys
  gen_sys_nonstrict.sing_rows = sg_strict;
  dim_type num_lines = num_rows(sg_strict);

  // Split the skeleton part.
  // Copies the sk-sys only if nnc_split and if both sides are non-empty.
  // Updates only `sat_strict': the copy is postponed.
  Ranges sk_ranges;
  auto [res_strict, res_nonstrict]
    = Q_split_skel(space_dim, num_lines, num_eq,
                   sk_strict, sk_nonstrict,
                   sk_ranges, beta_strict,
                   ns_strict, sat_c_strict);

  if (res_strict == Conv_Iter_Result::EMPTY ||
      res_nonstrict == Conv_Iter_Result::EMPTY)
    return { res_strict, res_nonstrict };

  // Non-skel handling is required only in NNC polyhedra.
  if (nnc_poly) {
    if (nnc_split) {
      split_non_skel(ns_strict, ns_nonstrict,
                     sk_strict, sk_nonstrict, sk_ranges,
                     sat_c_strict, sat_g_strict,
                     space_dim, num_lines, num_eq);
    } else {
      split_non_skel(ns_strict, ns_nonstrict, sk_strict, sk_ranges,
                     sat_c_strict, sat_g_strict);
    }
  }

  // Copy sat: all saturation info is shared.
  sat_c_nonstrict = sat_c_strict;
  // Note: we do NOT copy sat_g_strict into sat_g_nonstrict,
  // because these will be anyway rewritten by calls to simplify.
  if (not nnc_split)
    // Copy SK sys since it has been computed only once.
    sk_nonstrict = sk_strict;

  // Final SK update.
  Index_Partition<Range::NEG> sk_neg(sk_ranges);
  Index_Partition<Range::EQ> sk_eq(sk_ranges);
  Index_Partition<Range::POS> sk_pos(sk_ranges);
  // Update strict components
  if (nnc_split)
    points_become_closure_points(sk_eq, sk_strict);
  remove_rows(sk_neg, sk_strict, ns_strict, sat_c_strict);
  // Update nonstrict components
  remove_rows(sk_pos, sk_nonstrict, ns_nonstrict, sat_c_nonstrict);

  return { res_strict, res_nonstrict };
}

/**
 * 2-way rational split method.
 * Letting ph_in be the input value of ph_then, it updates
 *   ph_then = ph_in.add_con(beta)
 *   ph_else = ph_in.add_con(beta_complement)
 * Assumptions:
 *  - ph_then is non-empty and minimized;
 *  - beta is an inequality constraint;
 *  - if ph_then.topology() is CLOSED, then t is CLOSED too
 *    and beta is a non-strict inequality.
 */
void
Q_split_poly(Poly_Impl& ph_then, Poly_Impl& ph_else,
             const Con& beta, Topol t) {
  assert(ph_then.space_dim() == ph_else.space_dim());
  assert(beta.space_dim() <= ph_then.space_dim());
  assert(ph_then.is_minimized());
  assert(ph_else.is_empty());
  const bool nnc_poly = (ph_then.topology() == Topol::NNC);
  assert(t == Topol::CLOSED || nnc_poly);

  if (beta.is_inconsistent()) {
    using std::swap;
    swap(ph_then, ph_else);
    return;
  }
  if (beta.is_tautological() ||
      ph_then.is_empty() ||
      ph_then.space_dim() == 0)
    return;

  ph_else.set_status(Poly_Impl::Status::MINIMIZED);

  const dim_type space_dim = ph_then.space_dim();
  const dim_type num_eq = num_rows(ph_then.cs.sing_rows);

  assert(ph_else.gs.empty());
  // Copy original constraint system
  ph_else.cs = ph_then.cs;

  Con beta_else = complement_con(beta, t);
  const bool strict = beta.is_strict_inequality();
  const bool then_is_strict = (strict || t == Topol::CLOSED);

  // [non]strict suffixes are "real" only if t == Topol::NNC.
  auto& ph_strict = then_is_strict ? ph_then : ph_else;
  auto& ph_nonstrict = then_is_strict ? ph_else : ph_then;
  const auto& beta_strict = then_is_strict ? beta : beta_else;
  const auto& beta_nonstrict = then_is_strict ? beta_else : beta;
  auto& gens_strict = ph_strict.gs;
  auto& gens_nonstrict = ph_nonstrict.gs;
  auto& sat_c_strict = ph_strict.sat_c;
  auto& sat_c_nonstrict = ph_nonstrict.sat_c;
  auto& sat_g_strict = ph_strict.sat_g;
  auto& sat_g_nonstrict = ph_nonstrict.sat_g;
  auto& cons_strict = ph_strict.cs;
  auto& cons_nonstrict = ph_nonstrict.cs;

  // The "strict" side holds the original system.
  if (!then_is_strict) {
    using std::swap;
    swap(gens_strict, gens_nonstrict);
    swap(sat_c_strict, sat_c_nonstrict);
    swap(sat_g_strict, sat_g_nonstrict);
  }

  assert((t == Topol::NNC) == beta_strict.is_strict_inequality());
  const auto [res_strict, res_nonstrict]
    = Q_split_gen_sys(space_dim, num_eq,
                      gens_strict, gens_nonstrict,
                      sat_c_strict, sat_c_nonstrict,
                      sat_g_strict,
                      beta_strict, nnc_poly);
  // For rational splits, these are the only results allowed.
  assert((res_strict == Conv_Iter_Result::OK &&
          res_nonstrict == Conv_Iter_Result::OK)
         ||
         (res_strict == Conv_Iter_Result::EMPTY &&
          res_nonstrict == Conv_Iter_Result::REDUNDANT)
         ||
         (res_strict == Conv_Iter_Result::REDUNDANT &&
          res_nonstrict == Conv_Iter_Result::EMPTY));

  // If the strict side is empty, then `beta_nonstrict' is redundant,
  // so that ph_nonstrict is the same of input polyhedron ph_strict:
  // just swap them.
  if (res_strict == Conv_Iter_Result::EMPTY) {
    using std::swap;
    swap(ph_strict, ph_nonstrict);
    ph_strict.set_empty();
    return;
  }

  // Add the constraints (if not redundant).
  if (res_strict == Conv_Iter_Result::OK) {
    assert(res_nonstrict == Conv_Iter_Result::OK);
    cons_strict.add(beta_strict);
    cons_nonstrict.add(beta_nonstrict);
  }
  // Final simplification.
  if (res_strict == Conv_Iter_Result::EMPTY)
    ph_strict.set_empty();
  else
    simplify(space_dim, cons_strict, sat_c_strict, sat_g_strict,
             gens_strict);
  if (res_nonstrict == Conv_Iter_Result::EMPTY)
    ph_nonstrict.set_empty();
  else
    simplify(space_dim, cons_nonstrict, sat_c_nonstrict, sat_g_nonstrict,
             gens_nonstrict);
}

/******************** Integral split code ********************/

/**
 * Computes the split sk systems, copying the original one.
 * NB: sk_[then|else] is consistent only if result is not EMPTY.
 * If inconsistent, it is up to the caller to set the empty DD.
 */
Split_Result
Z_split_skel(dim_type space_dim, dim_type num_lines, dim_type num_eq,
             Gens& sk_then, Gens& sk_else,
             Ranges& sk_ranges_then, Ranges& sk_ranges_else,
             const Con& beta_then, const Con& beta_else,
             Sat& sat_then, Sat& sat_else) {
  assert(beta_then.is_nonstrict_inequality());
  // Note: sk_then, sat_then and cs_then are in-out params;
  // sk_else, sat_else and cs_else are out params.
  // We want to be lazy in populating the *_else params,
  // so as to factor common computations as much as possible.
  const Gens& sk_in = sk_then;
  const Sat& sat_in = sat_then;
  assert(sk_else.empty() && sat_else.empty());

  // Start processing the `then' component.
  Integers scal_prods_then;
  sp::compute_prods(beta_then, sk_in, scal_prods_then);
  sk_partition(scal_prods_then, sk_ranges_then);
  Index_Partition<Range::NEG> sk_neg_then(sk_ranges_then);
  Index_Partition<Range::EQ> sk_eq_then(sk_ranges_then);
  Index_Partition<Range::POS> sk_pos_then(sk_ranges_then);
  bool combine_then = not sk_neg_then.empty() && not sk_pos_then.empty();
  // Check for emptiness/redundancy.
  auto res_then = combine_then ? Conv_Iter_Result::OK
    : sk_neg_then.empty() ? Conv_Iter_Result::REDUNDANT
    : (sk_pos_then.empty() && has_no_eq_point(sk_ranges_then, sk_in)
       ? Conv_Iter_Result::EMPTY
       : Conv_Iter_Result::OK);

  // Now process `else' component, avoiding duplication of work.
  Integers scal_prods_else;
  if (combine_then) {
    // Need to perform a copy.
    scal_prods_else = scal_prods_then;
  } else {
    // Can perform a move.
    scal_prods_else = std::move(scal_prods_then);
  }

  auto Z_adjust_scal_prods
    = [](Integers& sps, const Gens& sk, const Integer& delta_inhomo) {
      for (auto i : index_range(sps)) {
        neg_assign(sps[i]);
        if (sk[i].is_point())
          add_mul_assign(sps[i], delta_inhomo, sk[i].divisor());
      }
    };

  // Do not recompute scalar products, "adjust" them.
  Integer delta = beta_then.inhomo_term() + beta_else.inhomo_term();
  Z_adjust_scal_prods(scal_prods_else, sk_in, delta);

  sk_partition(scal_prods_else, sk_ranges_else);
  Index_Partition<Range::NEG> sk_neg_else(sk_ranges_else);
  Index_Partition<Range::EQ> sk_eq_else(sk_ranges_else);
  Index_Partition<Range::POS> sk_pos_else(sk_ranges_else);
  bool combine_else = not sk_neg_else.empty() && not sk_pos_else.empty();

  auto res_else = combine_else ? Conv_Iter_Result::OK
    : sk_neg_else.empty() ? Conv_Iter_Result::REDUNDANT
    : (sk_pos_else.empty() && has_no_eq_point(sk_ranges_else, sk_in)
       ? Conv_Iter_Result::EMPTY
       : Conv_Iter_Result::OK);

  // Eager return in case both are empty.
  if (res_then == Conv_Iter_Result::EMPTY &&
      res_else == Conv_Iter_Result::EMPTY)
    return { res_then, res_else };

  // Needed when combining.
  const dim_type min_sat = space_dim - num_lines - 1;
  const dim_type max_sat = sat_in.num_cols() + num_eq;

  auto Z_adjust_sat = [](Ranges& sk_ranges, Sat& sat) {
    sat.add_cols(1);
    Index_Partition<Range::POS> sk_pos(sk_ranges);
    update_last_sat_column(sat, sk_pos);
  };

  auto Z_cleanup = [](Ranges& sk_ranges, Gens& sk, Sat& sat) {
    NS_Rows ns_dummy;
    Index_Partition<Range::NEG> sk_neg(sk_ranges);
    remove_rows(sk_neg, sk, ns_dummy, sat);
  };

  auto Z_combine = [min_sat, max_sat, Z_adjust_sat, Z_cleanup]
    (Ranges& sk_ranges, Gens& sk, Integers& sp, Sat& sat) {
    combine_sk_rows(sk_ranges, sk, sp, sat, false, min_sat, max_sat);
    Z_adjust_sat(sk_ranges, sat);
    Z_cleanup(sk_ranges, sk, sat);
  };

  if (combine_then && combine_else) {
    // Combine both.
    sk_else = sk_in;
    sat_else = sat_in;
    Z_combine(sk_ranges_then, sk_then, scal_prods_then, sat_then);
    Z_combine(sk_ranges_else, sk_else, scal_prods_else, sat_else);
    return { res_then, res_else };
  }

  if (combine_then) {
    assert(not combine_else);
    // To be computed before combining `then'
    if (res_else != Conv_Iter_Result::EMPTY) {
      sk_else = sk_in;
      sat_else = sat_in;
      if (res_else == Conv_Iter_Result::OK) {
        Z_adjust_sat(sk_ranges_else, sat_else);
        Z_cleanup(sk_ranges_else, sk_else, sat_else);
      }
    }
    // Now combine `then'
    Z_combine(sk_ranges_then, sk_then, scal_prods_then, sat_then);
    return { res_then, res_else };
  }

  if (combine_else) {
    assert(not combine_then);
    // To be computed before combining `else'
    if (res_then == Conv_Iter_Result::EMPTY) {
      // can avoid copy
      sk_else = std::move(sk_then);
      sat_else = std::move(sat_then);
    } else {
      // perform copy
      sk_else = sk_in;
      sat_else = sat_in;
      if (res_then == Conv_Iter_Result::OK) {
        Z_adjust_sat(sk_ranges_then, sat_then);
        Z_cleanup(sk_ranges_then, sk_then, sat_then);
      }
    }
    // Now combine `else'
    Z_combine(sk_ranges_else, sk_else, scal_prods_else, sat_else);
    return { res_then, res_else };
  }

  // No combine at all: they can't be both empty.
  if (res_else != Conv_Iter_Result::EMPTY) {
    if (res_then == Conv_Iter_Result::EMPTY) {
      // avoid deep copy
      sk_else = std::move(sk_then);
      sat_else = std::move(sat_then);
    } else {
      // need to copy
      sk_else = sk_then;
      sat_else = sat_then;
    }
  }
  if (res_then == Conv_Iter_Result::OK) {
    Z_adjust_sat(sk_ranges_then, sat_then);
    Z_cleanup(sk_ranges_then, sk_then, sat_then);
  }
  if (res_else == Conv_Iter_Result::OK) {
    Z_adjust_sat(sk_ranges_else, sat_else);
    Z_cleanup(sk_ranges_else, sk_else, sat_else);
  }
  return { res_then, res_else };
}

Split_Result
Z_split_gen_sys(dim_type space_dim, dim_type num_eq,
                Gen_Sys& gen_sys_then, Gen_Sys& gen_sys_else,
                Sat& sat_c_then, Sat& sat_c_else,
                const Con& beta_then, const Con& beta_else) {
  Gens& sg_then = gen_sys_then.sing_rows;
  Gens& sg_else = gen_sys_else.sing_rows;

  // Check if all singular rows saturate beta_then.
  const dim_type ex = check_sing_rows(beta_then, sg_then);
  if (ex < num_rows(sg_then)) {
    // NOTE: optimizable; here we just copy and work twice.
    gen_sys_else = gen_sys_then;
    sat_c_else = sat_c_then;
    Ranges dummy;
    process_violating_singular(ex, beta_then,
                               gen_sys_then, sat_c_then,
                               dummy, dummy);
    process_violating_singular(ex, beta_else,
                               gen_sys_else, sat_c_else,
                               dummy, dummy);
    return { Conv_Iter_Result::OK, Conv_Iter_Result::OK };
  }

  // Here all lines saturate the constraint.
  // Copy lines
  sg_else = sg_then;
  dim_type num_lines = num_rows(sg_then);

  // Split the skeleton part.
  Gens& sk_then = gen_sys_then.sk_rows;
  Gens& sk_else = gen_sys_else.sk_rows;
  Ranges sk_ranges_then, sk_ranges_else;
  return Z_split_skel(space_dim, num_lines, num_eq,
                      sk_then, sk_else,
                      sk_ranges_then, sk_ranges_else,
                      beta_then, beta_else,
                      sat_c_then, sat_c_else);
}

/**
 * 2-way integral split method.
 * Letting
 *  - ph_in be the input value of ph_then,
 *  - beta_then be the integral refinement of beta,
 *  - beta_else be the integral complement of beta_then
 * it updates
 *   ph_then = ph_in.add_con(beta_then)
 *   ph_else = ph_in.add_con(beta_else)
 * Assumptions:
 *  - ph_then.topology() is CLOSED
 *  - ph_then is non-empty and minimized;
 *  - beta is a (strict or non-strict) inequality constraint;
 */
void
Z_split_poly_on_ineq(Poly_Impl& ph_then, Poly_Impl& ph_else,
                     const Con& beta) {
  assert(ph_then.is_necessarily_closed());
  assert(ph_else.is_necessarily_closed());
  assert(ph_then.space_dim() == ph_else.space_dim());
  assert(beta.space_dim() <= ph_then.space_dim());
  assert(not beta.is_equality());
  assert(ph_then.is_minimized() && not ph_then.is_empty());
  assert(ph_else.is_empty());

  if (beta.is_inconsistent()) {
    using std::swap;
    swap(ph_then, ph_else);
    return;
  }
  if (beta.is_tautological() ||
      ph_then.is_empty() ||
      ph_then.space_dim() == 0)
    return;

  ph_else.set_status(Poly_Impl::Status::MINIMIZED);

  auto& cons_then = ph_then.cs;
  auto& cons_else = ph_else.cs;
  auto& gens_then = ph_then.gs;
  auto& gens_else = ph_else.gs;
  auto& sat_c_then = ph_then.sat_c;
  auto& sat_c_else = ph_else.sat_c;
  auto& sat_g_then = ph_then.sat_g;
  auto& sat_g_else = ph_else.sat_g;

  assert(gens_else.empty());

  const dim_type space_dim = ph_then.space_dim();
  const dim_type num_eq = num_rows(cons_then.sing_rows);

  auto [beta_then, beta_else] = integral_complement_cons(beta);
  assert(beta_then.is_nonstrict_inequality());
  assert(beta_else.is_nonstrict_inequality());

  // Note: in the integral case both results may be empty.
  auto [res_then, res_else]
    = Z_split_gen_sys(space_dim, num_eq,
                      gens_then, gens_else,
                      sat_c_then, sat_c_else,
                      beta_then, beta_else);

  // If needed, copy/move cons_then into cons_else.
  if (res_else != Conv_Iter_Result::EMPTY) {
    if (res_then == Conv_Iter_Result::EMPTY) {
      cons_else = std::move(cons_then);
      sat_g_else = std::move(sat_g_then);
    } else
      cons_else = cons_then;
  }

  switch (res_then) {
  case Conv_Iter_Result::OK:
    cons_then.add(beta_then);
    simplify(space_dim, cons_then, sat_c_then, sat_g_then, gens_then);
    break;
  case Conv_Iter_Result::EMPTY:
    ph_then.set_empty();
    break;
  case Conv_Iter_Result::REDUNDANT:
    // nothing to do
    break;
  }

  switch (res_else) {
  case Conv_Iter_Result::OK:
    cons_else.add(beta_else);
    simplify(space_dim, cons_else, sat_c_else, sat_g_else, gens_else);
    break;
  case Conv_Iter_Result::EMPTY:
    ph_else.set_empty();
    break;
  case Conv_Iter_Result::REDUNDANT:
    // nothing to do
    break;
  }
}

/**
 * 2-way integral split on equality constraint.
 * Letting
 *  - ph_in be the input value of ph_eq,
 *  - beta_eq := (expr == 0) be an equality constraint,
 *  - beta_lt be the integral refinement of (expr < 0),
 *  - beta_gt be the integral refinement of (expr > 0)
 * it updates
 *   ph_eq = ph_in.add_con(beta_eq)
 *   ph_ineq = ph_in.add_con(beta_lt), if ph_in and beta_gt are disjoint;
 *           = ph_in.add_con(beta_gt), if ph_in and beta_lt are disjoint;
 *           = ph_in,                  otherwise.
 * Assumptions:
 *  - ph_eq.topology() is CLOSED
 *  - ph_eq is non-empty and minimized;
 *  - ph_ineq is empty, CLOSED and same space dim as ph_eq
 *  - beta_eq is an equality constraint.
 */
void
Z_split_poly_on_eq(Poly_Impl& ph_eq, Poly_Impl& ph_ineq,
                   const Con& beta_eq) {
  assert(ph_eq.is_necessarily_closed());
  assert(ph_ineq.is_necessarily_closed());
  assert(ph_eq.space_dim() == ph_ineq.space_dim());
  assert(ph_eq.is_minimized() && not ph_eq.is_empty());
  assert(ph_ineq.is_empty());
  assert(beta_eq.is_equality());
  assert(beta_eq.space_dim() <= ph_eq.space_dim());

  if (is_integral_inconsistent(beta_eq)) {
    using std::swap;
    swap(ph_eq, ph_ineq);
    return;
  }
  if (beta_eq.is_tautological() ||
      ph_eq.is_empty() ||
      ph_eq.space_dim() == 0)
    return;

  auto rel_eq = ph_eq.relation_with(beta_eq);
  if (rel_eq.implies(Poly_Con_Rel::is_included()))
    return;
  if (rel_eq.implies(Poly_Con_Rel::is_disjoint())) {
    using std::swap;
    swap(ph_eq, ph_ineq);
    return;
  }

  auto [beta_lt, beta_gt] = integral_complement_eq(beta_eq);
  auto rel_lt = ph_eq.relation_with(beta_lt);
  if (rel_lt.implies(Poly_Con_Rel::is_included())) {
    using std::swap;
    swap(ph_eq, ph_ineq);
    return;
  }
  if (rel_lt.implies(Poly_Con_Rel::is_disjoint())) {
    ph_ineq = ph_eq;
    ph_eq.add_con(beta_eq);
    ph_ineq.add_con(beta_gt);
    return;
  }
  auto rel_gt = ph_eq.relation_with(beta_gt);
  assert(not rel_gt.implies(Poly_Con_Rel::is_included()));
  if (rel_gt.implies(Poly_Con_Rel::is_disjoint())) {
    ph_ineq = ph_eq;
    ph_eq.add_con(beta_eq);
    ph_ineq.add_con(beta_lt);
    return;
  }
  // Here we cannot refine ph_ineq
  ph_ineq = ph_eq;
  ph_eq.add_con(beta_eq);
}

void
Z_split_poly(Poly_Impl& ph_then, Poly_Impl& ph_else,
             const Con& beta) {
  if (beta.is_equality())
    Z_split_poly_on_eq(ph_then, ph_else, beta);
  else
    Z_split_poly_on_ineq(ph_then, ph_else, beta);
}

} // namespace detail
} // namespace pplite
